"""AI4 Metadata validator."""

import os

VERSIONS = {
    "1.0.0": os.path.join(os.path.dirname(__file__), "schemata/ai4-apps-v1.0.0.json"),
    "2.0.0": os.path.join(os.path.dirname(__file__), "schemata/ai4-apps-v2.0.0.json"),
}

LATEST_VERSION = "2.0.0"


def get_latest_schema() -> str:
    """Return the path to the latest schema."""
    return VERSIONS[LATEST_VERSION]


def get_schema(version: str) -> str:
    """Return the path to the schema for the given version."""
    return VERSIONS[version]


def get_latest_version() -> str:
    """Return the latest version."""
    return LATEST_VERSION


def get_all_versions() -> list[str]:
    """Return all available versions."""
    return VERSIONS.keys()



"""Unit tests for the validator module."""

import pathlib
import pytest

from ai4_metadata import utils
from ai4_metadata import validate


@pytest.fixture
def valid_schema_file():
    """Fixture for a valid schema file."""
    return open(pathlib.Path(__file__).parent / "../schemata/ai4-apps-v2.0.0.json", "r")


@pytest.fixture
def invalid_schema_file():
    """Fixture for an invalid schema file."""
    return None


@pytest.fixture
def valid_instance():
    """Fixture for a valid instance."""
    i = open(
        pathlib.Path(__file__).parent / "../../../instances/sample-v2.mods.json", "r"
    )
    return utils.load_json(i)


@pytest.fixture
def invalid_instance():
    """Fixture for an invalid instance."""
    return {"foo": "bar"}


def test_validator(valid_schema_file, valid_instance):
    """Test the validator."""
    assert validate.validate(valid_instance, valid_schema_file) is None


def test_validator_invalid_instance(valid_schema_file, invalid_instance):
    """Test the validator with an invalid instance."""
    with pytest.raises(Exception):  # noqa
        validate.validate(invalid_instance, valid_schema_file)


def test_validator_invalid_schema(invalid_schema_file, valid_instance):
    """Test the validator with an invalid schema."""
    with pytest.raises(Exception):  # noqa
        validate.validate(valid_instance, invalid_schema_file)


def test_validator_invallid_all(invalid_schema_file, invalid_instance):
    """Test the validator with an invalid schema and instance."""
    with pytest.raises(Exception):  # noqa
        validate.validate(invalid_instance, invalid_schema_file)



"""Main module for AI4 metadata validator."""

import argparse
import sys
import typing

from jsonschema import validators

import ai4_metadata
from ai4_metadata import utils


def validate(instance: object, schema_file: typing.TextIO) -> None:
    """Validate the schema."""
    schema = utils.load_json(schema_file)

    try:
        validator = validators.validator_for(schema)
        validator.check_schema(schema)
    except Exception as e:
        print(f"Error validating schema: {e}")
        raise

    validators.validate(instance, schema)


def main() -> None:
    """Validate the AI4 metadata schema via CLI."""
    parser = argparse.ArgumentParser(
        description=("AI4 application metadata (JSON-schema based) validator.")
    )

    version_group = parser.add_mutually_exclusive_group()

    version_group.add_argument(
        "--schema",
        metavar="SCHEMA_JSON",
        type=argparse.FileType("r"),
        help="AI4 application metadata schema file to use. "
        "If set, overrides --metadata-version.",
    )

    version_group.add_argument(
        "--metadata-version",
        metavar="VERSION",
        choices=ai4_metadata.get_all_versions(),
        default=ai4_metadata.get_latest_version(),
        help="AI4 application metadata version "
        f"(default: {ai4_metadata.get_latest_version()})",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Suppress output for valid instances",
    )

    parser.add_argument(
        "instance",
        metavar="METADATA_JSON",
        type=argparse.FileType("r"),
        nargs="+",
        help="AI4 application metadata file to validate.",
    )
    args = parser.parse_args()

    schema = args.schema or open(ai4_metadata.get_schema(args.metadata_version), "r")

    exit_code = 0
    for f in args.instance:
        try:
            instance = utils.load_json(f)
            validate(instance, schema)
        except Exception as e:
            print(f"Error validating instance: {e}")
            exit_code = 1
        else:
            if not args.quiet:
                print(f"{f.name} is valid for version {args.metadata_version}")

    sys.exit(exit_code)



"""Utility functions for the AI4 Metadata utils."""

import typing

import simplejson as json


def load_json(f: typing.TextIO) -> typing.Dict:
    """Load a JSON from the file f."""
    try:
        data = f.read()
        return json.loads(data)
    except json.JSONDecodeError as e:
        print(f"Error loading schema as JSON: {e}")
        raise



"""Migrate metadata from V1 to V2."""

import argparse
import collections
import datetime
import sys

import simplejson as json

import ai4_metadata
from ai4_metadata import utils
from ai4_metadata import validate


def migrate(v1_metadata: dict) -> collections.OrderedDict:
    """Try to migrate metadata from V1 to latest V2."""
    v2 = collections.OrderedDict()

    v2["metadata_version"] = ai4_metadata.get_latest_version()
    v2["title"] = v1_metadata.get("title")
    v2["summary"] = v1_metadata.get("summary")
    v2["description"] = " ".join(v1_metadata.get("description", []))
    v2["dates"] = {
        "created": v1_metadata.get("date_creation"),
        "updated": datetime.datetime.now().strftime("%Y-%m-%d"),
    }
    v2["links"] = {
        "source_code": v1_metadata.get("sources", {}).get("code"),
        "docker_image": v1_metadata.get("sources", {}).get("docker_registry_repo"),
    }
    v2["tags"] = v1_metadata.get("keywords", [])
    v2["tasks"] = []
    v2["categories"] = []
    v2["libraries"] = []

    # Now move things, if present, into links
    if v1_metadata.get("doi"):
        v2["links"]["doi"] = v1_metadata.get("doi")
    if v1_metadata.get("sources", {}).get("zenodo_doi"):
        v2["links"]["zenodo_doi"] = v1_metadata.get("sources", {}).get("zenodo_doi")
    if v1_metadata.get("sources", {}).get("pre_trained_weights"):
        v2["links"]["weights"] = v1_metadata.get("sources", {}).get(
            "pre_trained_weights"
        )
    if v1_metadata.get("sources", {}).get("ai4_template"):
        v2["links"]["ai4_template"] = v1_metadata.get("sources", {}).get("ai4_template")

    if v1_metadata.get("dataset_url"):
        v2["links"]["dataset_url"] = v1_metadata.get("dataset_url")
    if v1_metadata.get("training_files_url"):
        v2["links"]["training_files_url"] = v1_metadata.get("training_files_url")
    if v1_metadata.get("cite_url"):
        v2["links"]["cite_url"] = v1_metadata.get("cite_url")

    # Try to infer some some more information on libraries and categories
    kw = [k.lower() for k in v1_metadata.get("keywords", [])]
    if "tensorflow" in kw:
        v2["libraries"].append("TensorFlow")
    if "pytorch" in kw:
        v2["libraries"].append("PyTorch")
    if "keras" in kw:
        v2["libraries"].append("Keras")
    if "scikit-learn" in kw:
        v2["libraries"].append("Scikit-Learn")

    if "trainable" in kw:
        v2["categories"].append("AI4 trainable")
    if "inference" in kw:
        v2["categories"].append("AI4 inference")
        v2["categories"].append("AI4 pre trained")

    return v2


def main() -> None:
    """Convert metadata from V1 to latest V2."""
    parser = argparse.ArgumentParser(
        description=(
            "Migrate AI4 metadata from V1 to latest V2 (currently "
            f"{ai4_metadata.get_latest_version()})."
        )
    )

    parser.add_argument(
        "--output",
        "-o",
        metavar="OUTPUT_JSON",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Output file for migrated metadata, default is stdout.",
    )

    parser.add_argument(
        "instance",
        metavar="METADATA_JSON",
        type=argparse.FileType("r"),
        help="AI4 application metadata file to migrate.",
    )

    args = parser.parse_args()

    v1_metadata = utils.load_json(args.instance)
    v1_schema = open(ai4_metadata.get_schema("1.0.0"), "r")
    validate.validate(v1_metadata, v1_schema)
    print("V1 metadata is valid, continuing with migration...")

    # Migrate metadata
    v2_metadata = migrate(v1_metadata)
    v2_schema = open(ai4_metadata.get_schema(ai4_metadata.get_latest_version()), "r")
    validate.validate(v2_metadata, v2_schema)

    # Write out the migrated metadata
    json.dump(v2_metadata, args.output, indent=4)
